from abc import ABC, abstractmethod

from switcore.action.activity_router import PathResolver
from switcore.action.base_activity_handler import BaseHandler
from switcore.action.exceptions import UndefinedSubmitAction
from switcore.action.schemas import SwitRequest, BaseState, SwitResponse, UserActionType, QueryResource
from switcore.logger import get_logger
from switcore.type import DrawerHandler

logger = get_logger()


class ActivityHandlerABC(BaseHandler, ABC):
    def on_turn(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        if swit_request.user_action.type == UserActionType.view_actions_drop:
            response = self.on_view_actions_drop(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.view_actions_submit:
            response = self.on_view_actions_submit(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.right_panel_open:
            response = self.on_right_panel_open(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.view_actions_input:
            response = self.on_view_actions_input(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.view_actions_oauth_complete:
            response = self.on_view_actions_oauth_complete(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_chat_extension:
            response = self.on_user_commands_chat_extension(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_chat:
            response = self.on_user_commands_chat(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.presence_sync:
            response = self.on_presence_sync(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.view_actions_query:
            assert isinstance(swit_request.user_action.resource, QueryResource), "is not QueryResource"
            response = self.on_query(swit_request, state, swit_request.user_action.resource.value, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_task_extension:
            response = self.on_user_commands_task_extension(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_chat_commenting:
            response = self.on_user_commands_chat_commenting(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_context_menus_task:
            assert isinstance(swit_request.user_action.resource, dict), "is not dict"
            response = self.on_user_commands_context_menus_task(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_context_menus_message:
            response = self.on_user_commands_context_menus_message(swit_request, state, *args, **kwargs)
        elif swit_request.user_action.type == UserActionType.user_commands_context_menus_message_comment:
            response = self.on_user_commands_context_menus_message_comment(swit_request, state, *args, **kwargs)
        else:
            assert False, "undefined user_action type"

        return response

    @abstractmethod
    def on_right_panel_open(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_presence_sync(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_query(self, swit_request: SwitRequest, state: BaseState, query: str, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_chat(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_chat_extension(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_chat_commenting(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_context_menus_message(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_context_menus_message_comment(self, swit_request: SwitRequest,
                                                       state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_view_actions_drop(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    def on_view_actions_input(self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        return self.on_view_actions_submit(swit_request, state, *args, **kwargs)

    def on_view_actions_submit(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        user_action: str = swit_request.user_action.id
        action_path_resolver: PathResolver = PathResolver.from_combined(user_action)
        assert swit_request.current_view is not None, "current_view does not exist!!"
        view_key: str = PathResolver.from_combined(swit_request.current_view.view_id).key

        func_or_null: DrawerHandler | None = self._handlers.get(action_path_resolver.key, None)  # type: ignore
        view_ids_or_null: set[str] | None = self._view_actions.get(action_path_resolver.key, None)

        if func_or_null is None or view_ids_or_null is None:
            raise UndefinedSubmitAction(f"undefined submit action: {user_action}")

        if '*' not in view_ids_or_null and view_key not in view_ids_or_null:
            raise UndefinedSubmitAction(f"action is called in :{view_ids_or_null}")

        args = [swit_request, state, *action_path_resolver.paths, *args]
        response = func_or_null(*args, **kwargs)
        return response

    @abstractmethod
    def on_view_actions_oauth_complete(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_task_extension(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()

    @abstractmethod
    def on_user_commands_context_menus_task(
            self, swit_request: SwitRequest, state: BaseState, *args, **kwargs) -> SwitResponse:
        raise NotImplementedError()



import os

from fastapi import APIRouter, Depends, status, Request
from fastapi.responses import RedirectResponse

from switcore.auth.dependencies import get_oauth
from switcore.auth.oauth2 import OAuth2

router = APIRouter()


@router.get("/bot")
async def bot_authorization(
        request: Request,  # noqa: F841
        oauth: OAuth2 = Depends(get_oauth)
):
    return RedirectResponse(
        oauth.get_bot_authorization_url(),
        status.HTTP_307_TEMPORARY_REDIRECT
    )


@router.get("/user")
async def user_authorization(
        request: Request,  # noqa: F841
        oauth: OAuth2 = Depends(get_oauth)
):
    scopes: str | None = os.getenv('SCOPES', None)

    assert scopes is not None, "SCOPES is not set check .env file"

    return RedirectResponse(
        oauth.get_user_authorization_url(scopes),
        status.HTTP_307_TEMPORARY_REDIRECT
    )


"""
in order to add callback router url to your app you need to add this snippet to your app.py file

class BotAuthorizeCallback(OAuth2AuthorizeCallbackABC):
    async def process_access_token(_self, access_token: SwitToken, state: str | None) -> http_response:  # noqa

        self.assertEqual(access_token.access_token, expected["access_token"])
        self.assertEqual(access_token.expires_in, expected["expires_in"])
        self.assertEqual(access_token.refresh_token, expected["refresh_token"])
        self.assertEqual(access_token.scope, expected["scope"])
        self.assertEqual(access_token.token_type, expected["token_type"])

        return HTMLResponse(content="<script>window.close()</script>", status_code=status.HTTP_200_OK)


@self.app.get("/auth/callback/bot")
async def bot_oauth_callback(
        request: Request,
        code: str,
        state: str,
        oauth: OAuth2 = Depends(get_oauth)
):
    callback = BotAuthorizeCallback(
        client=oauth,
        redirect_url="test_base_url/auth/callback/bot",
    )

    return await callback(
        code=code,
        state=state,
    )
"""



import base64
import bz2
import json
from datetime import datetime
from enum import Enum
from typing import Any, Type
from typing import List

from pydantic import validator, root_validator

from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.button import Button
from switcore.ui.collection_entry import CollectionEntry
from switcore.ui.container import Container
from switcore.ui.datepicker import DatePicker
from switcore.ui.divider import Divider
from switcore.ui.file import File
from switcore.ui.header import Header, AttachmentHeader
from switcore.ui.html_frame import HtmlFrame
from switcore.ui.image import Image
from switcore.ui.image_grid import ImageGrid
from switcore.ui.info_card import InfoCard
from switcore.ui.input import Input
from switcore.ui.interactive_image import InteractiveImage
from switcore.ui.select import Select, Option, OptionGroup, NoOptionsReason
from switcore.ui.signIn_page import SignInPage
from switcore.ui.tabs import Tabs
from switcore.ui.text_paragraph import TextParagraph
from switcore.ui.textarea import Textarea


class UserInfo(SwitBaseModel):
    user_id: str
    organization_id: str


class UserPreferences(SwitBaseModel):
    language: str
    time_zone_offset: str
    color_theme: str


class Settings(SwitBaseModel):
    presence_sync: bool


class MessageResource(SwitBaseModel):
    resource_type: str
    id: str
    created_at: datetime
    edited_at: datetime | None
    content: str
    content_formatted: dict | None
    attachments: list[dict] | None
    files: list[dict] | None
    creator: dict
    settings: Settings | None = None

    @validator('edited_at', pre=True, allow_reuse=True)
    def parse_empty_string_to_none(cls, v):
        if v == '':
            return None
        return v


class SettingsResource(SwitBaseModel):
    resource_type: str
    settings: Settings


class QueryResource(SwitBaseModel):
    resource_type: str
    value: str


class UserActionType(str, Enum):
    right_panel_open = "right_panel_open"
    presence_sync = "presence_sync"
    user_commands_chat = "user_commands.extensions:chat"
    user_commands_chat_extension = "user_commands.chat_extension"
    user_commands_chat_commenting = "user_commands.extensions:chat_commenting"
    user_commands_context_menus_message = "user_commands.context_menus:message"
    user_commands_context_menus_message_comment = "user_commands.context_menus:message_comment"
    view_actions_drop = "view_actions.drop"
    view_actions_input = "view_actions.input"
    view_actions_query = "view_actions.query"
    view_actions_submit = "view_actions.submit"
    view_actions_oauth_complete = "view_actions.oauth_complete"
    user_commands_context_menus_task = "user_commands.context_menus:task"  # this action has a dict resource
    user_commands_task_extension = "user_commands.extensions:task"


class UserAction(SwitBaseModel):
    type: UserActionType
    id: str
    slash_command: str
    resource: MessageResource | SettingsResource | QueryResource | dict | None = None
    value: str | None = None

    class Config:
        smart_union = True

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        if isinstance(self.resource, dict):
            resource_type = self.resource.get('resource_type')
            if resource_type == 'query':
                self.resource = QueryResource(**self.resource)
            elif resource_type == 'settings.presence_sync':
                self.resource = SettingsResource(**self.resource)
            elif resource_type == 'message' or resource_type == 'message_comment':
                self.resource = MessageResource(**self.resource)


class Context(SwitBaseModel):
    workspace_id: str | None = None
    channel_id: str | None = None
    project_id: str | None = None
    task_id: str | None = None


ElementType = (CollectionEntry | Button | Divider | File | HtmlFrame | Input | Select
               | SignInPage | TextParagraph | Image | Textarea | Container | Tabs | DatePicker | InfoCard
               | ImageGrid | InteractiveImage)

AttachmentElementType = (CollectionEntry | InfoCard | Image | Divider | File | TextParagraph)


def contain_only_dict(elements_data: list[dict | ElementType]) -> bool:
    """
        if Body is initialized from swit_request, elements_data is list of dict
    """
    for element_data in elements_data:
        if isinstance(element_data, ElementType):
            return False

    return True


def get_element_type(element_data: dict) -> Type[ElementType]:
    element_type_str = element_data.get('type')
    if element_type_str == 'collection_entry':
        return CollectionEntry
    elif element_type_str == 'button':
        return Button
    elif element_type_str == 'divider':
        return Divider
    elif element_type_str == 'file':
        return File
    elif element_type_str == 'html_frame':
        return HtmlFrame
    elif element_type_str == 'text_input':
        return Input
    elif element_type_str == 'select':
        return Select
    elif element_type_str == 'sign_in_page':
        return SignInPage
    elif element_type_str == 'text':
        return TextParagraph
    elif element_type_str == 'textarea':
        return Textarea
    elif element_type_str == 'image':
        return Image
    elif element_type_str == 'container':
        return Container
    elif element_type_str == 'tabs':
        return Tabs
    elif element_type_str == 'datepicker':
        return DatePicker
    elif element_type_str == 'info_card':
        return InfoCard
    elif element_type_str == 'image_grid':
        return ImageGrid
    elif element_type_str == 'interactive_image':
        return InteractiveImage
    else:
        raise ValueError(f"Unknown element type: {element_type_str}")


class ElementMixin:
    def __init__(self, **data: Any) -> None:
        elements_data: list[dict | ElementType] = data.get('elements', [])

        if contain_only_dict(elements_data):
            _elements: List[ElementType] = []
            for element_data in elements_data:
                assert isinstance(element_data, dict), "element_data must be dict"
                element_type = get_element_type(element_data)
                element: ElementType = element_type(**element_data)
                _elements.append(element)
            data['elements'] = _elements

        super().__init__(**data)  # type: ignore

    def get_element_by_action_id(self, action_id: str) -> ElementType:
        """
        Get element by action_id if it exists, otherwise raise ValueError
        """
        elements: list[ElementType] = getattr(self, 'elements', [])
        for element in elements:  # type: ignore
            if getattr(element, 'action_id', None) is None:
                continue

            if element.action_id == action_id:  # type: ignore
                return element

        raise ValueError(f"Element with action_id: {action_id} not found")


class AttachmentBody(ElementMixin, SwitBaseModel):
    elements: list[AttachmentElementType]

    class Config:
        smart_union = True

    def __init__(self, **data: Any) -> None:
        """
        elements are one of  (CollectionEntry | InfoCard | Image | Divider | File | TextParagraph)
        """
        super().__init__(**data)


class Body(ElementMixin, SwitBaseModel):
    elements: list[ElementType]

    class Config:
        smart_union = True

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)


class Footer(SwitBaseModel):
    buttons: list[Button]


class ViewCallbackType(str, Enum):
    update = "views.update"
    initialize = "views.initialize"
    open = "views.open"
    push = "views.push"
    close = "views.close"


class AttachmentCallbackTypes(str, Enum):
    share_channel = "attachments.share.channel"
    share_new_task = "attachments.share.new_task"
    share_existing_task = "attachments.share.existing_task"


class SettingsCallbackTypes(str, Enum):
    settings_update = "settings.update"


class BotCallbackTypes(str, Enum):
    invite_prompt = "bot.invite_prompt"


class SuggestionsCallbackTypes(str, Enum):
    query_suggestions = "query.suggestions"


class SettingsResult(SwitBaseModel):
    success: bool = True
    error_message: str | None = None


class SuggestionsResult(SwitBaseModel):
    options: list[Option] | None = None
    option_groups: list[OptionGroup] | None = None
    no_options_reason: NoOptionsReason | None = None

    @root_validator
    def check_suggestions_result(cls, values):
        options = values.get('options')
        option_groups = values.get('option_groups')
        no_suggestions_reason = values.get('no_suggestions_reason')

        if options:
            if option_groups or no_suggestions_reason:
                raise ValueError('If options are set, option_groups and no_suggestions_reason should not be set')

        if option_groups:
            if options or no_suggestions_reason:
                raise ValueError('If option_groups are set, options and no_suggestions_reason should not be set')

        if no_suggestions_reason:
            if options or option_groups:
                raise ValueError('If no_suggestions_reason is set, options and option_groups should not be set')

        return values

    @validator('options')
    def validate_options_length(cls, v):
        if len(v) > 50:
            raise ValueError("options length should be less than 50")
        return v

    @validator('option_groups')
    def validate_option_groups_length(cls, v):
        if len(v) > 10:
            raise ValueError("option_groups length should be less than 10")
        return v


class DestinationTypes(str, Enum):
    channel = 'channel'
    project = 'project'


class Destination(SwitBaseModel):
    type: DestinationTypes
    id: str


class AttachmentView(SwitBaseModel):
    view_id: str
    state: str | bytes
    header: AttachmentHeader
    footer: Footer | None = None
    body: AttachmentBody


class View(SwitBaseModel):
    view_id: str
    state: str | bytes
    header: Header
    footer: Footer | None = None
    body: Body


class PlatformTypes(str, Enum):
    DESKTOP = 'Desktop'
    IOS = 'iOS'
    ANDROID = 'Android'


class SwitRequest(SwitBaseModel):
    platform: PlatformTypes
    time: datetime
    app_id: str
    user_info: UserInfo
    user_preferences: UserPreferences
    context: Context
    user_action: UserAction
    current_view: View | AttachmentView | None

    @validator('current_view', pre=True)
    def empty_dict_to_null(cls, v):
        if v == {}:
            return None
        return v


class SwitResponse(SwitBaseModel):
    callback_type: (ViewCallbackType | AttachmentCallbackTypes | SettingsCallbackTypes
                    | BotCallbackTypes | SuggestionsCallbackTypes)
    new_view: View | None = None
    attachments: list[AttachmentView] | None = None
    reference_view_id: str | None = None
    result: SettingsResult | SuggestionsResult | None = None
    destination: Destination | None = None

    class Config:
        smart_union = True


class BaseState(SwitBaseModel):
    autoincrement_id: int = 1

    @classmethod
    def from_bytes(cls, byte: bytes):
        d = json.loads(bz2.decompress(base64.b64decode(byte)).decode("utf-8"))
        return cls(**d)

    def to_bytes(self) -> bytes:
        return base64.b64encode(bz2.compress(self.json().encode("utf-8"), 1))



import datetime
import os

from fastapi import FastAPI

from switcore.action.async_activity_handler_abc import AsyncActivityHandlerABC
from switcore.action.schemas import SwitRequest, PlatformTypes, UserInfo, UserPreferences, Context, UserAction, \
    UserActionType, View, Body, BaseState, SwitResponse
from switcore.ui.header import Header


def create_fastapi_app():
    os.environ["SWIT_CLIENT_ID"] = "test_client_id"
    os.environ["SWIT_CLIENT_SECRET"] = "test_client_secret"
    os.environ["SWIT_SINGING_KEY"] = "test_singing_key"
    os.environ["APPS_ID"] = "test_apps_id"
    os.environ["DB_USERNAME"] = "test_db_username"
    os.environ["DB_PASSWORD"] = "test_db_password"
    os.environ["DB_NAME"] = "test_db_name"
    os.environ["ENV_OPERATION"] = "local"
    os.environ["BASE_URL"] = "test_base_url"
    os.environ["LOCALIZE_PROJECT_ID"] = "test_localize_project_id"
    os.environ["DB_HOST"] = "test_db_host"
    os.environ[
        "SCOPES"] = ("imap:write+imap:read+user:read+message:write+channel:read+workspace:read+project"
                     ":read+project:write+task:read+task:write")
    os.environ["BOT_REDIRECT_URL"] = "/auth/callback/bot"
    os.environ["USER_REDIRECT_URL"] = "/auth/callback/user"

    return FastAPI()


class ActivityHandler(AsyncActivityHandlerABC):

    def on_user_commands_task_extension(self, swit_request: SwitRequest, state: BaseState, *args,
                                        **kwargs) -> SwitResponse:  # type: ignore
        pass

    def on_user_commands_context_menus_task(self, swit_request: SwitRequest, state: BaseState, *args,
                                            **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_query(self, swit_request: SwitRequest, state: BaseState, query: str, *args,
                       **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_right_panel_open(self, swit_request: SwitRequest, state: BaseState, *args,
                                  **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_presence_sync(self, swit_request: SwitRequest, state: BaseState, *args,
                               **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_user_commands_chat(self, swit_request: SwitRequest, state: BaseState, *args,
                                    **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_user_commands_chat_extension(self, swit_request: SwitRequest, state: BaseState, *args,
                                              **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_user_commands_chat_commenting(self, swit_request: SwitRequest, state: BaseState, *args,
                                               **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_user_commands_context_menus_message(self, swit_request: SwitRequest, state: BaseState, *args,
                                                     **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_user_commands_context_menus_message_comment(self, swit_request: SwitRequest, state: BaseState, *args,
                                                             **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_view_actions_drop(self, swit_request: SwitRequest, state: BaseState, *args,
                                   **kwargs) -> SwitResponse:  # type: ignore
        pass

    async def on_view_actions_oauth_complete(self, swit_request: SwitRequest, state: BaseState, *args,
                                             **kwargs) -> SwitResponse:  # type: ignore
        pass


def create_submit_swit_request(view_id: str, action_id: str):
    return SwitRequest(
        platform=PlatformTypes.DESKTOP,
        time=datetime.datetime.now(),
        app_id="test_app_id",
        user_info=UserInfo(
            user_id="test_user_id",
            organization_id="test_organization_id"
        ),
        user_preferences=UserPreferences(
            language="ko",
            time_zone_offset="+0900",
            color_theme="light"
        ),
        context=Context(
            workspace_id="test_workspace_id",
            channel_id="test_channel_id"
        ),
        user_action=UserAction(
            type=UserActionType.view_actions_submit,
            id=action_id,
            slash_command="test_slash_command",
        ),
        current_view=View(
            view_id=view_id,
            state="state",
            header=Header(
                title="test_title",
            ),
            body=Body(
                elements=[]
            ),
        ))


def create_query_swit_request():
    _json = {'time': '2023-12-27T06:12:22.498Z', 'app_id': '230824012947339SP3AX', 'platform': 'Desktop',
             'context': {'workspace_id': '', 'channel_id': '', 'project_id': '', 'task_id': '', 'message': None,
                         'task': None, 'approval_request_id': ''},
             'user_action': {'type': 'view_actions.query', 'id': 'example14_server_query', 'slash_command': '',
                             'resource': {'resource_type': 'query', 'value': 'q'}},
             'user_preferences': {'language': 'en', 'time_zone_offset': '+0900', 'color_theme': 'light'},
             'user_info': {'user_id': '23042402102425L72QG7', 'organization_id': '2209220405404NAp2M9'}}

    return SwitRequest(**_json)


def create_task_swit_request():
    _json = {'time': '2023-12-27T06:12:22.498Z', 'app_id': '230824012947339SP3AX', 'platform': 'Desktop',
             'context': {'workspace_id': '', 'channel_id': '', 'project_id': '', 'task_id': '', 'message': None,
                         'task': None, 'approval_request_id': ''},
             'user_action': {'type': 'user_commands.context_menus:task',
                             'id': 'example14_server_query', 'slash_command': '',
                             'resource': {
                                 "resource_type": "task",
                                 "id": "23051106260551ZOVCPS",
                                 "parent_task_id": "200302045745598izpK",
                                 "created_at": "2022-07-07T12:41:18Z",
                                 "edited_at": "2022-07-08T07:22:18Z",
                                 "title": "Developers documentation",
                                 "content": "Hello, World!",
                                 "content_formatted": {
                                     "type": "rich_text",
                                     "elements": [
                                         {
                                             "type": "rt_section",
                                             "elements": [
                                                 {
                                                     "type": "rt_text",
                                                     "content": "Hello, World!"
                                                 }
                                             ]
                                         }
                                     ]
                                 },
                                 "period": {
                                     "start_time": "2023-02-14T05:01:00.452591Z",
                                     "due_time": "2023-02-14T05:01:00.452591Z",
                                     "include_time": True
                                 },
                                 "priority": "highest",
                                 "color_label": "red",
                                 "assignees": [
                                     {
                                         "id": "220103011810x7bqTRZ"
                                     }
                                 ],
                                 "collaborators": [
                                     {
                                         "id": "220103011810x7bqTRZ"
                                     }
                                 ],
                                 "tags": [
                                     {
                                         "name": "Open API"
                                     }
                                 ],
                                 "status": {
                                     "id": "220707214090V110qhk",
                                     "name": "Done",
                                     "type": "not_started"
                                 },
                                 "bucket": {
                                     "id": "230213021820Zc9NjbT",
                                     "name": "Help Center"
                                 }
                             }},
             'user_preferences': {'language': 'en', 'time_zone_offset': '+0900', 'color_theme': 'light'},
             'user_info': {'user_id': '23042402102425L72QG7', 'organization_id': '2209220405404NAp2M9'}}

    return SwitRequest(**_json)



import json  # noqa: F401
import unittest

from switcore.action.schemas import SwitRequest, UserActionType, \
    QueryResource
from tests.utils import create_submit_swit_request, create_query_swit_request, create_task_swit_request


class SwitViewResponseTest(unittest.TestCase):

    def test_swit_submit_request(self):
        swit_request: SwitRequest = create_submit_swit_request("right_panel", "test_action_id01")
        self.assertEqual(swit_request.user_action.type, UserActionType.view_actions_submit)

    def test_swit_query_request(self):
        swit_request: SwitRequest = create_query_swit_request()
        self.assertEqual(swit_request.user_action.type, UserActionType.view_actions_query)
        self.assertTrue(isinstance(swit_request.user_action.resource, QueryResource))

    def test_swit_task_request(self):
        swit_request: SwitRequest = create_task_swit_request()
        self.assertEqual(swit_request.user_action.type, UserActionType.user_commands_context_menus_task.value)
        self.assertTrue(isinstance(swit_request.user_action.resource, dict))
        assert isinstance(swit_request.user_action.resource, dict)
        self.assertTrue("id" in swit_request.user_action.resource)
        self.assertTrue("parent_task_id" in swit_request.user_action.resource)
        self.assertTrue("title" in swit_request.user_action.resource)
        self.assertTrue("status" in swit_request.user_action.resource)
        self.assertTrue("assignees" in swit_request.user_action.resource)



import unittest

from switcore.ui.interactive_image import InteractiveImage
from switcore.ui.image_grid import ImageGrid


class InteractiveImageTest(unittest.TestCase):

    def test_valid_interactive_image01(self):
        expected: dict = {
            "type": "image_grid",
            "images": [
                {
                    "alt": "image test",
                    "type": "interactive_image",
                    "action_id": "action_01",
                    "draggable": False,
                    "image_url": "https://www.example.com/images/sample.png"
                }
            ],
            "column_count": 2
        }

        image = InteractiveImage(
            image_url="https://www.example.com/images/sample.png",
            alt="image test",
            action_id="action_01"
        )
        grid = ImageGrid(
            images=[image],
            column_count=2
        )
        actual: dict = grid.dict(exclude_none=True)
        self.assertEqual(expected, actual)

    def test_valid_interactive_image02(self):
        expected: dict = {
            "type": "image_grid",
            "images": [
                {
                    "alt": "image test1",
                    "type": "interactive_image",
                    "action_id": "action_01",
                    "draggable": False,
                    "image_url": "https://www.example.com/images/sample1.png"
                },
                {
                    "alt": "image test2",
                    "type": "interactive_image",
                    "action_id": "action_02",
                    "draggable": True,
                    "image_url": "https://www.example.com/images/sample2.png"
                },
            ],
            "column_count": 3
        }

        image1 = InteractiveImage(
            image_url="https://www.example.com/images/sample1.png",
            alt="image test1",
            action_id="action_01"
        )
        image2 = InteractiveImage(
            image_url="https://www.example.com/images/sample2.png",
            alt="image test2",
            action_id="action_02",
            draggable=True
        )
        grid = ImageGrid(
            images=[image1, image2],
            column_count=3
        )

        actual: dict = grid.dict(exclude_none=True)

        self.assertEqual(expected, actual)



import unittest

from switcore.ui.collection_entry import CollectionEntry, TextSection
from switcore.ui.image import Image
from switcore.ui.text_paragraph import TextParagraph


class CollectionEntryTest(unittest.TestCase):

    def test_valid_collection_entry01(self):
        collection_entry: CollectionEntry = CollectionEntry(
            text_sections=[
                TextSection(
                    text=TextParagraph(
                        content="test content"
                    ),
                    metadata_items=[
                        Image(image_url="./assets/builder_logo.png")
                    ])
            ]
        )
        expected: dict = {
            'type': 'collection_entry',
            'draggable': False,
            'text_sections': [
                {
                    'metadata_items': [
                        {
                            'image_url': './assets/builder_logo.png',
                            'type': 'image'
                        }
                    ],
                    'text': {
                        'content': 'test content',
                        'markdown': False,
                        'type': 'text'}
                }
            ],
            'vertical_alignment': 'top'
        }

        # print(json.dumps(expected, indent=4))
        self.assertEqual(expected, collection_entry.dict(exclude_none=True))

    def test_valid_collection_entry02(self):
        collection_entry: CollectionEntry = CollectionEntry(
            start_section=Image(
                image_url="./assets/builder_logo.png",
            ),
            text_sections=[
                TextSection(
                    text=TextParagraph(
                        content="test content"
                    ),
                )
            ]
        )
        expected: dict = {
            'type': 'collection_entry',
            'start_section': {
                'image_url': './assets/builder_logo.png',
                'type': 'image'
            },
            'draggable': False,
            'text_sections': [
                {
                    'text': {
                        'content': 'test content',
                        'markdown': False,
                        'type': 'text'}
                }
            ],
            'vertical_alignment': 'top'
        }

        # print(json.dumps(expected, indent=4))
        self.assertEqual(expected, collection_entry.dict(exclude_none=True))



from sqlalchemy import Column, String, Integer, DateTime, func
from sqlalchemy.ext.declarative import declarative_base

base_model = declarative_base()


class BaseModel(base_model):
    __abstract__ = True

    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    created = Column(DateTime, default=func.now())
    updated = Column(DateTime, default=func.now(), onupdate=func.now())


class App(BaseModel):
    __tablename__ = 'app'

    access_token: str = Column(String(500))
    refresh_token: str = Column(String(50))
    iss: str = Column(String(50))
    cmp_id: str = Column(String(50), nullable=True)
    apps_id: str = Column(String(50))


class User(BaseModel):
    __tablename__ = 'user'

    swit_id: str = Column(String(100))
    access_token: str = Column(String(500))
    refresh_token: str = Column(String(50))



from switcore.pydantic_base_model import SwitBaseModel


class SearchInput(SwitBaseModel):
    type: str = 'search_input'
    action_id: str
    placeholder: str | None
    value: str | None
    disabled: bool = False



from typing import Optional

from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.Icon import Icon
from switcore.ui.element_components import StaticAction


class Button(SwitBaseModel):
    id: Optional[str] = None
    type: str = "button"
    label: str | None = None
    style: str | None = None
    disabled: bool | None = None
    action_id: str | None = None
    static_action: StaticAction | None = None
    icon: Icon | None = None



import unittest

from switcore.ui.button import Button
from switcore.ui.element_components import WriteToClipboard, OpenLink, CloseView


class StaticActionTest(unittest.TestCase):

    def test_clipboard_copy(self):
        button = Button(
            label="button",
            static_action=WriteToClipboard(
                content="Copied Successfully",
            )
        )
        self.assertTrue(isinstance(button.static_action, WriteToClipboard))
        expected = {
            'label': 'button',
            'static_action': {'action_type': 'write_to_clipboard', 'content': 'Copied Successfully'},
            'type': 'button'
        }
        self.assertEqual(expected, button.dict(exclude_none=True))

    def test_open_link(self):
        button = Button(
            label="button",
            static_action=OpenLink(
                link_url="https://www.google.com",
            )
        )
        expected = {
            'label': 'button',
            'static_action': {'action_type': 'open_link', 'link_url': 'https://www.google.com'},
            'type': 'button'
        }
        self.assertEqual(expected, button.dict(exclude_none=True))

    def test_close_view(self):
        button = Button(
            label="button",
            static_action=CloseView()
        )
        self.assertTrue(isinstance(button.static_action, CloseView))

        expected = {
            'label': 'button',
            'static_action': {'action_type': 'close_view'},
            'type': 'button'
        }
        self.assertEqual(expected, button.dict(exclude_none=True))



import unittest

from switcore.ui.element_components import OpenLink, Tag, TagStyle, TagColorTypes, TagShapeTypes
from switcore.ui.image import Image
from switcore.ui.select import Select, Option, OptionGroup, SelectQuery, NoOptionsReason


class SelectTest(unittest.TestCase):

    def test_valid_select01(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1"
                ),
                Option(
                    label="test label2",
                    action_id="action_id2"
                ),
            ]
        )
        expected = {
            'type': 'select',
            'multiselect': False,
            'trigger_on_input': True,
            'options': [
                {
                    'label': 'test label1',
                    'action_id': 'action_id1'
                },
                {
                    'label': 'test label2',
                    'action_id': 'action_id2'
                }
            ],
            'option_groups': []
        }
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_valid_select02(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1",
                    static_action=OpenLink(
                        link_url="https://www.google.com"
                    )
                ),
                Option(
                    label="test label2",
                    action_id="action_id2"
                ),
            ]
        )
        expected = {
            'type': 'select',
            'multiselect': False,
            'options': [{'action_id': 'action_id1',
                         'label': 'test label1',
                         'static_action': {'link_url': 'https://www.google.com',
                                           'action_type': 'open_link'}},
                        {'action_id': 'action_id2', 'label': 'test label2'}],
            'option_groups': [],
            'trigger_on_input': True,
        }
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_valid_select03(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="action_id1",
                    tag=Tag(content="tag1", style=TagStyle(color=TagColorTypes.danger))
                ),
                Option(
                    label="test label2",
                    action_id="action_id2",
                    tag=Tag(content="tag1", style=TagStyle(color=TagColorTypes.primary, shape=TagShapeTypes.rounded))
                ),
            ]
        )

        expected = {
            'type': 'select',
            'multiselect': False,
            'options': [{'action_id': 'action_id1', 'label': 'test label1',
                         'tag': {'type': 'tag', 'content': 'tag1',
                                 'style': {'color': 'danger', 'shape': 'rectangular'}}},
                        {'action_id': 'action_id2', 'label': 'test label2',
                         'tag': {'type': 'tag', 'content': 'tag1', 'style': {'color': 'primary', 'shape': 'rounded'}}}],
            'option_groups': [],
            'trigger_on_input': True,
        }
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_options_select(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    icon=Image(
                        image_url="https://www.google.com",
                    ),
                    action_id="action_id1",
                    tag=Tag(content="tag1"),
                ),
                Option(
                    label="test label2",
                    action_id="action_id2",
                    tag=Tag(content="tag2"),
                ),
                Option(
                    label="test label3",
                    icon=Image(
                        image_url="https://www.google.com",
                    ),
                    action_id="action_id3",
                ),
            ],
            query=SelectQuery(
                query_server=True,
                action_id="action_id"
            )
        )

        expected: dict = {
            'type': 'select',
            'multiselect': False,
            'options': [
                {'action_id': 'action_id1', 'icon': {'type': 'image', 'image_url': 'https://www.google.com'},
                 'label': 'test label1', 'tag': {'type': 'tag', 'content': 'tag1'}},
                {'action_id': 'action_id2', 'label': 'test label2',
                 'tag': {'type': 'tag', 'content': 'tag2'}},
                {'action_id': 'action_id3', 'icon': {'type': 'image', 'image_url': 'https://www.google.com'},
                 'label': 'test label3'}
            ],
            'option_groups': [],
            'trigger_on_input': True,
            'query': {'query_server': True, 'disabled': False, 'action_id': 'action_id'}
        }
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_group_options_select(self):
        select = Select(
            trigger_on_input=True,
            option_groups=[
                OptionGroup(
                    label="test group1",
                    options=[
                        Option(
                            label="test label1",
                            icon=Image(
                                image_url="https://www.google.com",
                            ),
                            tag=Tag(content="tag1"),
                            action_id="action_id1",
                        ),
                        Option(
                            label="test label2",
                            action_id="action_id2",
                            tag=Tag(content="tag2"),
                        )
                    ],
                ),
                OptionGroup(
                    label="test group2",
                    options=[
                        Option(
                            label="test label3",
                            icon=Image(
                                image_url="https://www.google.com",
                            ),
                            action_id="action_id3",
                        )
                    ],
                ),

            ],
            query=SelectQuery(
                query_server=True,
                action_id="action_id"
            )
        )

        expected: dict = {
            'type': 'select',
            'multiselect': False,
            'options': [],
            'option_groups': [
                {'label': 'test group1', 'options': [
                    {'action_id': 'action_id1', 'icon': {'type': 'image', 'image_url': 'https://www.google.com'},
                     'label': 'test label1', 'tag': {'type': 'tag', 'content': 'tag1'}},
                    {'action_id': 'action_id2', 'label': 'test label2',
                     'tag': {'type': 'tag', 'content': 'tag2'}}
                ]},
                {'label': 'test group2', 'options': [
                    {'action_id': 'action_id3', 'icon': {'type': 'image', 'image_url': 'https://www.google.com'},
                     'label': 'test label3'}
                ]}
            ],
            'trigger_on_input': True,
            'query': {'query_server': True, 'disabled': False, 'action_id': 'action_id'}
        }
        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_mixed_select(self):
        with self.assertRaises(ValueError):
            Select(
                trigger_on_input=True,
                options=[
                    Option(
                        label="test label1",
                        action_id="mixed_action_id1",
                    ),
                    Option(
                        label="test label2",
                        action_id="mixed_action_id2"
                    ),
                ],
                option_groups=[
                    OptionGroup(label="test group1",
                                options=[Option(label="test label1", action_id="mixed_action_id3")]),
                    OptionGroup(label="test group2",
                                options=[Option(label="test label2", action_id="mixed_action_id4")]),
                ],
                query=SelectQuery(
                    action_id="query_action_id_3"
                )
            )

    def test_filter_select(self):
        select = Select(
            trigger_on_input=True,
            options=[
                Option(
                    label="test label1",
                    action_id="filter_action_1",
                ),
                Option(
                    label="test label2",
                    action_id="filter_action_2"
                ),
            ],
            query=SelectQuery(
                query_server=False,
                action_id="query_action_id_4"
            )
        )

        expected: dict = {
            'type': 'select',
            'multiselect': False,
            'options': [{'action_id': 'filter_action_1', 'label': 'test label1'},
                        {'action_id': 'filter_action_2', 'label': 'test label2'}],
            'option_groups': [],
            'trigger_on_input': True,
            'query': {'query_server': False, 'disabled': False, 'action_id': 'query_action_id_4'}

        }

        self.assertEqual(expected, select.dict(exclude_none=True))

    def test_select_option_groups_length(self):
        with self.assertRaises(ValueError):
            Select(
                trigger_on_input=True,
                option_groups=[
                    OptionGroup(label="test group1",
                                options=[]),
                    OptionGroup(label="test group2",
                                options=[Option(label="test label2", action_id="mixed_action_id4")]),
                ],
                query=SelectQuery(
                    action_id="query_action_id_3"
                )
            )

    def test_select_no_options_reason(self):
        select = Select(
            trigger_on_input=True,
            no_options_reason=NoOptionsReason(
                message="test message"
            ),
            query=SelectQuery(
                query_server=False,
                action_id="query_action_id_3"
            )
        )

        expected: dict = {
            'type': 'select',
            'multiselect': False,
            'options': [],
            'option_groups': [],
            'trigger_on_input': True,
            'no_options_reason': {'message': 'test message'},
            'query': {'query_server': False, 'disabled': False, 'action_id': 'query_action_id_3'}
        }

        self.assertEqual(expected, select.dict(exclude_none=True))



from enum import Enum

from switcore.pydantic_base_model import SwitBaseModel


class TextareaSizeTypes(str, Enum):
    small = "small"
    medium = "medium"
    large = "large"


class Textarea(SwitBaseModel):
    type: str = "textarea"
    action_id: str
    placeholder: str | None
    value: str | None
    height: TextareaSizeTypes = TextareaSizeTypes.small
    disabled: bool = False



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.button import Button
from switcore.ui.datepicker import DatePicker
from switcore.ui.select import Select


class Container(SwitBaseModel):
    type: str = "container"
    elements: list[Select | Button | DatePicker]

    class Config:
        smart_union = True



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.element_components import StaticAction


class SelectItem(SwitBaseModel):
    """
        A reusable object that represents an item that can be selected.
        It is used across various components like View, AttachmentView, Tabs, and Select.
    """
    label: str
    action_id: str
    static_action: StaticAction | None = None



from httpx import Response

from switcore.api.channel.schemas import Channel
from switcore.api.helpers import sort_by_name
from switcore.httpclient import CustomHTTPClient


@sort_by_name
def get_sync_all_channels_recursive(
        http_client: CustomHTTPClient,
        workspace_id: str,
        offset: str | None = None  # type: ignore
) -> list[Channel]:
    params: dict = {
        "limit": 50,
        "workspace_id": workspace_id,
        "type": "gen,dfl"
    }

    if offset:
        params["offset"] = offset

    res = http_client.api_get('/v1/api/channel.list', params=params)

    data: dict = res.json()
    data = data.get('data', {})

    channels_data: list[dict] = data.get('channels', [])
    offset: str | None = data.get('offset', None)

    if len(channels_data) == 0:
        return []

    channels = [Channel(**channel) for channel in channels_data]
    next_channels = get_sync_all_channels_recursive(http_client, workspace_id, offset)
    return channels + next_channels


def get_sync_channel_user_list(
        http_client: CustomHTTPClient,
        channel_id: str,
) -> Response:
    params: dict = {
        "channel_id": channel_id,
    }

    return http_client.api_get('/v1/api/channel.user.list', params=params)


def get_sync_channel(
        http_client: CustomHTTPClient,
        channel_id: str) -> Channel | None:
    params: dict = {
        "id": channel_id,
    }

    res = http_client.api_get('/v1/api/channel.info', params=params)

    data: dict = res.json()
    data = data.get('data', {})

    channel_data: dict | None = data.get('channel', None)

    if channel_data is None:
        return None

    return Channel(**channel_data)



import os

import httpx
from httpx import Response
from sqlalchemy.orm import Session

from switcore.auth.models import User
from switcore.auth.repository import AppRepository, UserRepository
from switcore.auth.schemas import SwitToken


class CustomAsyncHTTPClient(httpx.AsyncClient):
    def __init__(self,
                 base_url: str,
                 access_token: str,
                 refresh_token: str,
                 session: Session,
                 refresh_token_url,
                 swit_id_or_null: str | None = None,
                 app_id_or_null: str | None = None,
                 cmp_id_or_null: str | None = None,
                 *args, **kwargs):

        assert ((swit_id_or_null is None) ^ (app_id_or_null is None)), \
            "Either swit_id or app_id should be provided, but not both."

        if swit_id_or_null:
            assert cmp_id_or_null is None, "cmp_id를 넣어주세요"

        if app_id_or_null:
            assert cmp_id_or_null, "cmp_id를 넣어주세요"

        kwargs['timeout'] = 3.0
        super().__init__(*args, **kwargs)
        self.base_url = base_url
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.swit_id_or_null = swit_id_or_null
        self.app_id_or_null = app_id_or_null
        self.cmp_id_or_null = cmp_id_or_null
        self.refresh_token_url = refresh_token_url
        self.session = session
        self.is_user = swit_id_or_null is not None

    def add_auth_header(self):
        return {
            "Authorization": f"Bearer {self.access_token}",
        }

    async def refresh(self):
        client_id: str | None = os.getenv('SWIT_CLIENT_ID', None)
        client_secret: str | None = os.getenv('SWIT_CLIENT_SECRET', None)

        assert client_id is not None, "check SWIT_CLIENT_ID in env!!"
        assert client_secret is not None, "check SWIT_CLIENT_SECRET in env!!"

        data = {
            "grant_type": "refresh_token",
            "client_id": client_id,
            "client_secret": client_secret,
            "refresh_token": self.refresh_token,
        }

        url = self.build_url(self.refresh_token_url)

        response = await self.post(
            url, data=data, headers={'Content-Type': 'application/x-www-form-urlencoded'})
        swit_token: SwitToken = SwitToken(**response.json())
        if self.is_user:
            assert self.swit_id_or_null is not None, "swit_id should be provided"

            user: User = UserRepository(self.session).update_token(
                self.swit_id_or_null,
                swit_token.access_token,
                swit_token.refresh_token
            )
            self.access_token = user.access_token
            self.refresh_token = user.refresh_token
        else:
            assert self.cmp_id_or_null is not None, "cmp_id should be provided"
            assert self.app_id_or_null is not None, "app_id should be provided"

            app = AppRepository(self.session).update_token(
                self.cmp_id_or_null,
                self.app_id_or_null,
                swit_token.access_token,
                swit_token.refresh_token
            )
            self.access_token = app.access_token
            self.refresh_token = app.refresh_token

    def build_url(self, path):
        return f"{self.base_url}{path}"

    async def api_get(self, path, **kwargs):
        url = self.build_url(path)
        return await self.get(url, headers=self.add_auth_header(), **kwargs)

    async def api_post(self, path, data, **kwargs):
        url = self.build_url(path)
        return await self.post(url, headers=self.add_auth_header(), json=data, **kwargs)

    def is_refresh_request(self, args: tuple):
        return len(args) == 1 and args[0] == self.build_url(self.refresh_token_url)

    async def request(self, method: str, *args, **kwargs) -> Response:
        if self.is_refresh_request(args):  # is refresh token
            return await super().request(method, *args, **kwargs)
        else:
            response: Response = await super().request(method, *args, **kwargs)
            if response.status_code == 401:
                await self.refresh()
                kwargs['headers'] = self.add_auth_header()
                response = await super().request(method, *args, **kwargs)
                assert response.status_code != 401, "refresh token failed"
            return response



import json
import os

from httpx import Response

from switcore.async_httpclient import CustomAsyncHTTPClient
from switcore.httpclient import CustomHTTPClient
from switcore.logger import get_logger

logger = get_logger()


async def create_message(
        http_client: CustomAsyncHTTPClient,
        channel_id: str,
        content: dict,
        attachments: list[dict]
):
    app_id: str | None = os.getenv('APPS_ID', None)

    assert app_id is not None, "check APPS_ID in env!!"

    data = {
        'channel_id': channel_id,
        'body_type': 'json_string',
        'content': json.dumps(content),
        'app_id': app_id,
    }

    if len(attachments) > 0:
        data['attachments'] = attachments  # type: ignore

    try:
        response: Response = await http_client.api_post(
            path="/v1/api/message.create",
            data=data
        )
    except Exception as e:
        raise e

    if response.status_code == 200:
        logger.info(f"successful payload is {data}")
    else:
        logger.info(f"failed payload is {data} {response.status_code} {response.text}")
    return response


def create_sync_message(
        http_client: CustomHTTPClient,
        channel_id: str,
        content: dict,
        attachments: list[dict]
):
    app_id: str | None = os.getenv('APPS_ID', None)

    assert app_id is not None, "check APPS_ID in env!!"

    data = {
        'channel_id': channel_id,
        'body_type': 'json_string',
        'content': json.dumps(content),
        'app_id': app_id,
    }

    if len(attachments) > 0:
        data['attachments'] = attachments  # type: ignore

    try:
        response: Response = http_client.api_post(
            path="/v1/api/message.create",
            data=data
        )
    except Exception as e:
        raise e

    if response.status_code == 200:
        logger.info(f"successful payload is {data}")
    else:
        logger.info(f"failed payload is {data} {response.status_code} {response.text}")
    return response



import hmac
import hashlib
from time import time


class SignatureVerifier:
    def __init__(self, signing_key: str, max_delay: int = 60 * 5, secret_version: str = "s0="):
        self.signing_key = signing_key.encode()
        self.max_delay = max_delay
        self.secret_version = secret_version

    def is_valid(self, body: str, timestamp_or_null: str | None, signature_or_null: str | None) -> bool:
        if None in [body, timestamp_or_null, signature_or_null]:
            return False

        assert signature_or_null is not None, "signature_or_null is None"
        assert timestamp_or_null is not None, "timestamp_or_null is None"

        if self._is_timestamp_valid(timestamp_or_null):
            calculated_signature = self._generate_signature(body, timestamp_or_null)
            return hmac.compare_digest(calculated_signature, signature_or_null)
        return False

    def _is_timestamp_valid(self, timestamp: str) -> bool:
        return abs(time() - int(timestamp)) <= self.max_delay

    def _generate_signature(self, body: str, timestamp: str) -> str:
        base_string = f"swit:{timestamp}:{body}"
        signature = hmac.new(self.signing_key, base_string.encode(), hashlib.sha256)
        return self.secret_version + signature.hexdigest()



from pydantic import Field

from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.interactive_image import InteractiveImage


class ImageGrid(SwitBaseModel):
    type: str = "image_grid"
    images: list[InteractiveImage]
    column_count: int = Field(..., ge=2, le=3)



import unittest

from switcore.ui.element_components import OpenOauthPopup
from switcore.ui.interactive_image import InteractiveImage


class InteractiveImageTest(unittest.TestCase):

    def test_valid_interactive_image01(self):
        expected: dict = {
            "type": "interactive_image",
            "image_url": "https://swit.io/assets/images/home/features/device",
            "alt": "image test",
            "action_id": "action_01",
            "static_action": {
                "action_type": "open_oauth_popup",
                "link_url": "https://example.com"
            },
            "draggable": False
        }

        image = InteractiveImage(
            image_url="https://swit.io/assets/images/home/features/device",
            alt="image test",
            action_id="action_01",
            static_action=OpenOauthPopup(
                link_url="https://example.com"
            )
        )
        actual: dict = image.dict(exclude_none=True)
        self.assertEqual(expected, actual)

    def test_valid_interactive_image02(self):
        expected: dict = {
            "type": "interactive_image",
            "image_url": "https://swit.io/assets/images/home/features/device",
            "action_id": "action_01",
            "draggable": True
        }

        image = InteractiveImage(
            image_url="https://swit.io/assets/images/home/features/device",
            action_id="action_01",
            draggable=True
        )
        actual: dict = image.dict(exclude_none=True)
        self.assertEqual(expected, actual)



from enum import Enum

from switcore.pydantic_base_model import SwitBaseModel


class InputVariant(str, Enum):
    text = 'text'
    select = 'select'


class Input(SwitBaseModel):
    type: str = 'text_input'
    action_id: str
    placeholder: str | None
    trigger_on_input: bool = False
    value: str | None
    variant: InputVariant = InputVariant.text



import os

from switcore.constants import Environment


def get_swit_openapi_base_url() -> str:
    env: str | None = os.getenv('ENV_OPERATION', None)

    assert env is not None, "ENV_OPERATION is not set check .env file"

    if env == Environment.LOCAL or env == Environment.PROD:
        return 'https://openapi.swit.io'
    elif env == Environment.EXPRESS:
        return 'https://openapi.swit.express'
    elif env == Environment.DEV:
        return 'https://openapi2.swit.dev'
    else:
        raise ValueError(f"Invalid ENV_OPERATION: {env}")



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.Icon import Icon
from switcore.ui.button import Button


class IntegratedService(SwitBaseModel):
    icon: Icon


class SignInPage(SwitBaseModel):
    id: str | None = None
    type: str = "sign_in_page"
    integrated_service: IntegratedService
    title: str
    description: str
    button: Button



from sqlalchemy.orm import Session

from switcore.auth.exception import NotFoundException
from switcore.auth.models import User, App


class RepositoryBase:
    def __init__(self, session: Session):
        self.session = session


class AppRepository(RepositoryBase):
    def create(
            self,
            access_token: str,
            refresh_token: str,
            iss: str,
            apps_id: str,
            cmp_id: str | None
    ) -> App:
        token = App(  # type: ignore
            access_token=access_token,
            refresh_token=refresh_token,
            iss=iss,
            apps_id=apps_id,
            cmp_id=cmp_id
        )
        self.session.add(token)
        self.session.commit()
        return token

    def get_or_create(self, access_token: str, refresh_token: str, iss: str, apps_id: str, cmp_id: str) -> App:
        try:
            app = self.get_by_cmp_id_and_apps_id(cmp_id=cmp_id, apps_id=apps_id)
        except NotFoundException:
            app = self.create(
                access_token=access_token,
                refresh_token=refresh_token,
                iss=iss,
                apps_id=apps_id,
                cmp_id=cmp_id
            )
        return app

    def install(self, access_token: str, refresh_token: str, iss: str, apps_id: str, cmp_id: str) -> App:
        try:
            app = self.get_by_cmp_id_and_apps_id(cmp_id=cmp_id, apps_id=apps_id)
            self.session.delete(app)
            self.session.commit()
        except NotFoundException:
            pass

        app = self.create(
            access_token=access_token,
            refresh_token=refresh_token,
            iss=iss,
            apps_id=apps_id,
            cmp_id=cmp_id
        )

        return app

    def get_by_cmp_id_and_apps_id(self, cmp_id: str, apps_id: str) -> App:
        """
        :raises AppNotFoundException:
        """
        app_or_null: App | None = self.session.query(App).filter(App.cmp_id == cmp_id,
                                                                 App.apps_id == apps_id).first()  # type: ignore
        if app_or_null is None:
            raise NotFoundException(f"App with cmp_id: {cmp_id} and apps_id: {apps_id} not found")
        return app_or_null

    def update_token(self, cmp_id: str, apps_id: str, access_token: str, refresh_token: str) -> App:
        app = self.get_by_cmp_id_and_apps_id(cmp_id=cmp_id, apps_id=apps_id)
        app.access_token = access_token
        app.refresh_token = refresh_token
        self.session.commit()
        return app


class UserRepository(RepositoryBase):
    def create(self, swit_id: str, access_token: str, refresh_token: str) -> User:
        user = User(  # type: ignore
            swit_id=swit_id,
            access_token=access_token,
            refresh_token=refresh_token
        )
        self.session.add(user)
        self.session.commit()
        return user

    def get_or_create(self, swit_id: str, access_token: str, refresh_token: str):
        try:
            user = self.get_by_swit_id(swit_id=swit_id)
        except NotFoundException:
            user = self.create(
                swit_id=swit_id,
                access_token=access_token,
                refresh_token=refresh_token
            )
        return user

    def get_by_swit_id(self, swit_id: str) -> User:
        """
        :raises UserNotFoundException:
        """
        user_or_null: User | None = self.session.query(User).filter(User.swit_id == swit_id).first()  # type: ignore
        if user_or_null is None:
            raise NotFoundException(detail="User not found")
        return user_or_null

    def update_token(self, swit_id: str, access_token: str, refresh_token: str) -> User:
        user: User = self.get_by_swit_id(swit_id=swit_id)
        user.access_token = access_token
        user.refresh_token = refresh_token
        self.session.commit()
        return user

    def delete(self, swit_id: str) -> None:
        user = self.get_by_swit_id(swit_id)
        self.session.delete(user)
        self.session.commit()



import inspect
from typing import Callable

from switcore.api.channel.schemas import Channel
from switcore.api.workspace.schemas import Workspace


def sort_by_name(func: Callable) -> Callable:
    def sync_wrapper(*args, **kwargs) -> list[Workspace | Channel]:
        arr: list[Workspace | Channel] = func(*args, **kwargs)
        return sort_and_return(arr)

    async def async_wrapper(*args, **kwargs) -> list[Workspace | Channel]:
        arr: list[Workspace | Channel] = await func(*args, **kwargs)
        return sort_and_return(arr)

    def sort_and_return(arr: list[Workspace | Channel]) -> list[Workspace | Channel]:
        if len(arr) > 0 and hasattr(arr[0], "name"):
            arr.sort(key=lambda element: getattr(element, "name"))
        return arr

    if inspect.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper



from switcore.api.user.schemas import SwitUser
from switcore.async_httpclient import CustomAsyncHTTPClient
from switcore.httpclient import CustomHTTPClient


async def get_me(http_client: CustomAsyncHTTPClient) -> SwitUser | None:
    res = await http_client.api_get('/v1/api/user.info')

    data: dict = res.json()
    data = data.get('data', {})

    user_data: dict | None = data.get('user', None)

    if user_data is None:
        return None

    return SwitUser(**user_data)


def get_sync_me(http_client: CustomHTTPClient) -> SwitUser | None:
    res = http_client.api_get('/v1/api/user.info')

    data: dict = res.json()
    data = data.get('data', {})

    user_data: dict | None = data.get('user', None)

    if user_data is None:
        return None

    return SwitUser(**user_data)



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.item import Item


class InfoCard(SwitBaseModel):
    type: str = 'info_card'
    items: list[Item]
    action_id: str | None = None
    draggable: bool = False



from switcore.api.helpers import sort_by_name
from switcore.api.workspace.schemas import Workspace
from switcore.async_httpclient import CustomAsyncHTTPClient


@sort_by_name
async def get_all_workspaces_recursive(
        http_client: CustomAsyncHTTPClient,
        offset: str | None = None  # type: ignore
) -> list[Workspace]:
    params: dict = {
        "limit": 50,
    }

    if offset:
        params["offset"] = offset

    res = await http_client.api_get('/v1/api/workspace.list', params=params)

    data: dict = res.json()
    data = data.get('data', {})

    workspaces_data: list[dict] = data.get('workspaces', [])
    offset: str | None = data.get('offset', None)

    if len(workspaces_data) == 0:
        return []

    workspaces = [Workspace(**workspace) for workspace in workspaces_data]
    next_workspaces = await get_all_workspaces_recursive(http_client, offset)
    return workspaces + next_workspaces


async def get_workspace(
        http_client: CustomAsyncHTTPClient,
        workspace_id: str) -> Workspace | None:
    params: dict = {
        "id": workspace_id,
    }

    res = await http_client.api_get('/v1/api/workspace.info', params=params)

    data: dict = res.json()
    data = data.get('data', {})

    workspaces_data: dict | None = data.get('workspace', None)

    if workspaces_data is None:
        return None

    return Workspace(**workspaces_data)



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.element_components import StaticAction


class InteractiveImage(SwitBaseModel):
    type: str = "interactive_image"
    image_url: str
    alt: str | None = None
    action_id: str | None = None
    static_action: StaticAction | None = None
    draggable: bool = False



import unittest

from switcore.action.activity_router import PathResolver


class PathResolverTest(unittest.TestCase):
    def test_combined_path(self):
        path_resolver: PathResolver = PathResolver("test_action_id01", ["a"])
        self.assertEqual(str(path_resolver), "test_action_id01/a")
        self.assertEqual(path_resolver.id, "test_action_id01")
        self.assertEqual(path_resolver.paths, ["a"])

    def test_from_combined01(self):
        path_resolver: PathResolver = PathResolver.from_combined("test_action_id01/a/b")
        self.assertEqual(str(path_resolver), "test_action_id01/a/b")
        self.assertEqual(path_resolver.id, "test_action_id01")
        self.assertEqual(path_resolver.paths, ["a", "b"])

    def test_from_combined02(self):
        path_resolver: PathResolver = PathResolver.from_combined("test_action_id01/a/b/1")
        self.assertEqual(str(path_resolver), "test_action_id01/a/b/1")
        self.assertEqual(path_resolver.id, "test_action_id01")
        self.assertEqual(path_resolver.paths, ["a", "b", 1])

    def test_convert_int(self):
        path_resolver: PathResolver = PathResolver("test_action_id01", [1, "a"])
        self.assertEqual(str(path_resolver), "test_action_id01/1/a")
        self.assertEqual(path_resolver.paths, [1, "a"])

    def test_escape01(self):
        path_resolver: PathResolver = PathResolver("test_escape_action_id/escape", ["a"])
        self.assertEqual(path_resolver.combined_path, "test_escape_action_id%2Fescape/a")
        self.assertEqual(path_resolver.id, "test_escape_action_id/escape")
        self.assertEqual(path_resolver.paths, ["a"])

        path_resolver: PathResolver = PathResolver("test_escape_action_id/escape", ["a/b"])
        self.assertEqual(path_resolver.combined_path, "test_escape_action_id%2Fescape/a%2Fb")
        self.assertEqual(path_resolver.id, "test_escape_action_id/escape")
        self.assertEqual(path_resolver.paths, ["a/b"])

        path_resolver: PathResolver = PathResolver("test_action_id01/aa/bb", ["a/b"])
        self.assertEqual(path_resolver.combined_path, "test_action_id01%2Faa%2Fbb/a%2Fb")
        self.assertEqual(path_resolver.id, "test_action_id01/aa/bb")
        self.assertEqual(path_resolver.paths, ["a/b"])



import json  # noqa: F401
import unittest

from switcore.action.schemas import SwitResponse, ViewCallbackType, View, Body, AttachmentCallbackTypes, AttachmentView, \
    AttachmentBody, SuggestionsCallbackTypes, SuggestionsResult, NoOptionsReason
from switcore.ui.divider import Divider
from switcore.ui.element_components import Tag, TagStyle, TagColorTypes, TagShapeTypes
from switcore.ui.header import Header, AttachmentHeader
from switcore.ui.image import Image
from switcore.ui.select import Option, OptionGroup
from switcore.ui.text_paragraph import TextParagraph


class SwitViewResponseTest(unittest.TestCase):

    def test_swit_response_view(self):
        body = Body(
            elements=[TextParagraph(content="test content"), Divider()],
        )
        swit_response = SwitResponse(
            callback_type=ViewCallbackType.update,
            new_view=View(
                view_id="test01",
                state="test state",
                header=Header(title="this is Header"),
                body=body,
            )
        )
        expected: dict = {
            'callback_type': ViewCallbackType.update,
            'new_view': {
                'view_id': 'test01',
                'state': "test state",
                'header': {'title': 'this is Header'},
                'body': {
                    'elements': [
                        {
                            'content': 'test content',
                            'markdown': False,
                            'type': 'text'
                        },
                        {
                            'type': 'divider'
                        }
                    ],
                }
            }
        }
        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_attachments(self):
        swit_response = SwitResponse(
            callback_type=AttachmentCallbackTypes.share_channel,
            attachments=[AttachmentView(
                view_id="test01",
                state="test state",
                header=AttachmentHeader(
                    title="this is Header",
                    app_id="test app id",
                    icon=Image(
                        image_url="https://example.com/image.png",
                    )
                ),
                body=AttachmentBody(
                    elements=[TextParagraph(content="test content")],
                ),
            )]
        )
        expected: dict = {
            'callback_type': AttachmentCallbackTypes.share_channel,
            'attachments': [{
                'view_id': 'test01',
                'state': "test state",
                'header': {
                    'title': 'this is Header',
                    'app_id': 'test app id',
                    'icon': {
                        'type': 'image',
                        'image_url': 'https://example.com/image.png'
                    }
                },
                'body': {
                    'elements': [
                        {
                            'content': 'test content',
                            'markdown': False,
                            'type': 'text'
                        }
                    ],
                }
            }]
        }
        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_query_suggestion01(self):
        swit_response = SwitResponse(
            callback_type=SuggestionsCallbackTypes.query_suggestions,
            result=SuggestionsResult(
                options=[Option(
                    label="Search 성공!",
                    action_id="success_option_action_id",
                )],
            )
        )

        expected: dict = {
            'callback_type': SuggestionsCallbackTypes.query_suggestions,
            'result': {
                'options': [{
                    'label': 'Search 성공!',
                    'action_id': 'success_option_action_id'
                }]
            }
        }
        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_query_suggestion02(self):
        options: list[Option] = [
            Option(
                label="test label1",
                icon=Image(
                    image_url="https://files.swit.dev/webhook_logo.png",
                ),
                tag=Tag(content="no style"),
                action_id="test_action_id1"
            ),
            Option(
                label="test label2",
                action_id="test_action_id2",
                tag=Tag(content="red rounded", style=TagStyle(color=TagColorTypes.danger, shape=TagShapeTypes.rounded)),
            ),
            Option(
                label="test label3",
                icon=Image(
                    image_url="https://files.swit.dev/webhook_logo.png",
                ),
                action_id="test_action_id3",
            ),
        ]

        swit_response = SwitResponse(
            callback_type=SuggestionsCallbackTypes.query_suggestions,
            result=SuggestionsResult(
                options=options,
            )
        )

        expected: dict = {
            'callback_type': SuggestionsCallbackTypes.query_suggestions.value,
            'result': {
                'options': [
                    {
                        'label': 'test label1',
                        'icon': {
                            'type': 'image',
                            'image_url': 'https://files.swit.dev/webhook_logo.png'
                        },
                        'tag': {
                            'type': 'tag',
                            'content': 'no style'
                        },
                        'action_id': 'test_action_id1'
                    },
                    {
                        'label': 'test label2',
                        'tag': {
                            'type': 'tag',
                            'content': 'red rounded',
                            'style': {
                                'color': 'danger',
                                'shape': 'rounded'
                            }
                        },
                        'action_id': 'test_action_id2'
                    },
                    {
                        'label': 'test label3',
                        'icon': {
                            'type': 'image',
                            'image_url': 'https://files.swit.dev/webhook_logo.png'
                        },
                        'action_id': 'test_action_id3'
                    }
                ]
            }
        }

        # convert expected json to json format
        expected_json = json.dumps(expected)

        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_query_suggestions03(self):
        with self.assertRaises(ValueError):
            SwitResponse(
                callback_type=SuggestionsCallbackTypes.query_suggestions,
                result=SuggestionsResult(
                    options=[Option(
                        label="Search 성공!",
                        action_id="success_option_action_id",
                    )],
                    option_groups=[
                        OptionGroup(
                            label="test group",
                            options=[Option(
                                label="test label",
                                action_id="test action id",
                            )]
                        )
                    ]
                )
            )

    def test_swit_response_query_suggestion05(self):
        swit_response = SwitResponse(
            callback_type=SuggestionsCallbackTypes.query_suggestions,
            result=SuggestionsResult(
                option_groups=[
                    OptionGroup(
                        label="test group1",
                        options=[Option(
                            label="test label1",
                            action_id="test action id1",
                        )]
                    ),
                    OptionGroup(
                        label="test group2",
                        options=[Option(
                            label="test label2",
                            action_id="test action id2",
                        )]
                    ),
                ]
            )
        )

        expected: dict = {
            'callback_type': SuggestionsCallbackTypes.query_suggestions,
            'result': {
                'option_groups': [
                    {
                        'label': 'test group1',
                        'options': [{
                            'label': 'test label1',
                            'action_id': 'test action id1'
                        }]
                    },
                    {
                        'label': 'test group2',
                        'options': [{
                            'label': 'test label2',
                            'action_id': 'test action id2'
                        }]
                    }
                ]
            }
        }
        # convert expected json to json format
        expected_json = json.dumps(expected)

        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_query_suggestion06(self):
        swit_response = SwitResponse(
            callback_type=SuggestionsCallbackTypes.query_suggestions,
            result=SuggestionsResult(
                no_options_reason=NoOptionsReason(
                    message="test message",
                )
            )
        )

        expected: dict = {
            'callback_type': SuggestionsCallbackTypes.query_suggestions,
            'result': {
                'no_options_reason': {
                    'message': 'test message'
                }
            }
        }

        # convert expected json to json format
        expected_json = json.dumps(expected)

        self.assertEqual(expected, swit_response.dict(exclude_none=True))

    def test_swit_response_task_attachments(self):
        swit_response = SwitResponse(
            callback_type=AttachmentCallbackTypes.share_new_task,
            attachments=[AttachmentView(
                view_id="test01",
                state="test state",
                header=AttachmentHeader(
                    title="this is Header",
                    app_id="test app id",
                    icon=Image(
                        image_url="https://example.com/image.png",
                    )
                ),
                body=AttachmentBody(
                    elements=[TextParagraph(content="test content")],
                ),
            )]
        )
        expected: dict = {
            'callback_type': AttachmentCallbackTypes.share_new_task,
            'attachments': [{
                'view_id': 'test01',
                'state': "test state",
                'header': {
                    'title': 'this is Header',
                    'app_id': 'test app id',
                    'icon': {
                        'type': 'image',
                        'image_url': 'https://example.com/image.png'
                    }
                },
                'body': {
                    'elements': [
                        {
                            'content': 'test content',
                            'markdown': False,
                            'type': 'text'
                        }
                    ],
                }
            }]
        }
        self.assertEqual(expected, swit_response.dict(exclude_none=True))



from datetime import datetime

from pydantic import BaseModel


class Channel(BaseModel):
    created: datetime
    description: str
    host_id: str
    id: str
    is_archived: bool
    is_member: bool
    is_prev_chat_visible: bool
    is_private: bool
    is_starred: bool
    name: str
    type: str



import time
import unittest

from switcore.auth.signature_verification import SignatureVerifier


class SignatureVerifierTest(unittest.TestCase):
    def setUp(self):
        self.signing_key = "secret_key"
        self.request_body = "test_body"
        self.timestamp = str(int(time.time()))
        self.signature_verifier = SignatureVerifier(self.signing_key)
        self.signature = self.signature_verifier._generate_signature(self.request_body, self.timestamp)

    def test_verification_success(self):
        self.assertTrue(self.signature_verifier.is_valid(self.request_body, self.timestamp, self.signature))

    def test_verification_failure_due_to_time(self):
        expired_timestamp = str(int(time.time()) - 60 * 6)
        self.assertFalse(self.signature_verifier.is_valid(self.request_body, expired_timestamp, self.signature))

    def test_verification_failure_due_to_invalid_signing_secret(self):
        invalid_secret_key = "invalid_secret_key"
        invalid_signature_verifier = SignatureVerifier(invalid_secret_key)
        self.assertFalse(invalid_signature_verifier.is_valid(self.request_body, self.timestamp, self.signature))



from enum import Enum


class Environment(str, Enum):
    LOCAL = "local"
    DEV = "dev"
    EXPRESS = "express"
    PROD = "prod"

    @property
    def is_debug(self) -> bool:
        return self in (self.LOCAL, self.EXPRESS, self.DEV)



import datetime
import os
from abc import ABC, abstractmethod
from typing import Any
from urllib.parse import urlencode

import httpx
import jwt
from pydantic import BaseModel
from starlette.responses import Response

from switcore.auth.exception import InvalidOauthStateException
from switcore.auth.schemas import SwitToken


class OAuth2(BaseModel):
    client_id: str
    client_secret: str
    base_url: str
    authorize_endpoint: str
    access_token_endpoint: str
    refresh_token_endpoint: str
    bot_redirect_url: str
    user_redirect_url: str
    request_headers: dict[str, str] = {
        "content-type": "application/x-www-form-urlencoded",
        "Accept": "application/json",
    }

    def get_bot_authorization_url(self, state: str | None = None) -> str:
        bot_scopes_or_null: str | None = os.getenv("BOT_SCOPES", None)

        scope = "app:install"
        if bot_scopes_or_null is not None:
            scope = scope + "+" + bot_scopes_or_null

        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "scope": scope,
            "redirect_uri": self.base_url + self.bot_redirect_url,
        }

        if state is not None:
            params["state"] = state

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    def get_user_authorization_url(self, scope: str, state: str | None = None) -> str:
        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "scope": scope,
            "redirect_uri": self.base_url + self.user_redirect_url,
        }

        if state is not None:
            params["state"] = state

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    async def get_authorization_url(
            self,
            redirect_uri: str,
            scope: str,
            state: str | None = None,
            **extras_params,
    ) -> str:
        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": redirect_uri,
            "scope": scope,
        }

        if state is not None:
            params["state"] = state

        if extras_params is not None:
            params = {**params, **extras_params}

        return f"{self.authorize_endpoint}?{urlencode(params)}"

    async def get_access_token(self, code: str, redirect_uri: str) -> SwitToken:
        async with httpx.AsyncClient() as client:
            data = {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": redirect_uri,
                "client_id": self.client_id,
                "client_secret": self.client_secret,
            }

            response = await client.post(
                self.access_token_endpoint,
                data=data,
                headers=self.request_headers,
            )

            data: dict[str, Any] = response.json()

            if response.status_code >= 400:
                raise ValueError(data)

            return SwitToken(**data)

    async def refresh_token(self, refresh_token: str) -> SwitToken:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.refresh_token_endpoint,
                data={
                    "grant_type": "refresh_token",
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "refresh_token": refresh_token,
                },
                headers=self.request_headers,
            )

            data: dict[str, Any] = response.json()

            if response.status_code >= 400:
                raise ValueError(data)

            return SwitToken(**data)

    async def get_secure_authorization_url(
            self,
            redirect_uri: str,
            scope: str,
            callback_payload: dict | None = None,
            **extras_params) -> str:

        if callback_payload is None:
            callback_payload = {}

        callback_payload["exp"] = datetime.datetime.utcnow() + datetime.timedelta(minutes=5)

        swit_singing_key: str | None = os.getenv('SWIT_SINGING_KEY', None)

        assert swit_singing_key is not None, "SWIT_SINGING_KEY is not set check .env file"

        encoded_jwt = jwt.encode(callback_payload, swit_singing_key, algorithm="HS256")

        return await self.get_authorization_url(redirect_uri, scope, encoded_jwt, **extras_params)


class OAuth2AuthorizeCallbackABC(ABC):
    def __init__(self, client: OAuth2, redirect_url: str) -> None:
        self.client = client
        self.redirect_url = redirect_url

    @abstractmethod
    async def process_access_token(self, access_token: SwitToken, state: str | None) -> Response:
        pass

    async def _get_access_token(self, code: str) -> SwitToken:
        return await self.client.get_access_token(code, self.redirect_url)

    async def __call__(
            self,
            code: str,
            state: str | None,
            error: str | None = None,
    ) -> Response:
        access_token = await self._get_access_token(code)
        return await self.process_access_token(access_token, state)


class SecureOAuth2AuthorizeCallbackABC(ABC):
    def __init__(self, client: OAuth2, redirect_url: str) -> None:
        self.client = client
        self.redirect_url = redirect_url

    @abstractmethod
    async def process_access_token(self, access_token: SwitToken, callback_payload: dict) -> Response:
        pass

    async def _get_access_token(self, code: str) -> SwitToken:
        return await self.client.get_access_token(code, self.redirect_url)

    async def __call__(
            self,
            code: str,
            state: str,
            error: str | None = None,
    ) -> Response:
        swit_singing_key: str | None = os.getenv('SWIT_SINGING_KEY', None)

        assert swit_singing_key is not None, "SWIT_SINGING_KEY is not set check .env file"

        try:
            callback_payload = jwt.decode(state, swit_singing_key, algorithms=["HS256"])
        except Exception as e:
            raise InvalidOauthStateException(str(e))

        access_token = await self._get_access_token(code)
        return await self.process_access_token(access_token, callback_payload)



import unittest

from switcore.action.schemas import Body
from switcore.ui.button import Button
from switcore.ui.container import Container
from switcore.ui.datepicker import DatePicker
from switcore.ui.select import Select, Option


class ContainerTest(unittest.TestCase):

    def test_valid01(self):
        body: Body = Body(
            elements=[
                Container(
                    elements=[
                        DatePicker(
                            action_id="action_01",
                            trigger_on_input=False,
                        ),
                        Button(
                            label="Button 1",
                            style="primary_filled"
                        )
                    ]

                )]
        )

        expected = {
            "elements": [
                {
                    "type": "container",
                    "elements": [
                        {
                            "type": "datepicker",
                            "action_id": "action_01",
                            'trigger_on_input': False,
                        },
                        {
                            "type": "button",
                            "label": "Button 1",
                            "style": "primary_filled"
                        }
                    ]
                }
            ]
        }

        self.assertEqual(expected, body.dict(exclude_none=True))

    def test_valid02(self):
        select1 = Select(
            placeholder="Select an option",
            trigger_on_input=True,
            options=[
                Option(
                    label="Option 1",
                    action_id="option1"
                ),
                Option(
                    label="Option 2",
                    action_id="option2"
                ),
            ]
        )

        select2 = Select(
            placeholder="Select an option",
            trigger_on_input=True,
            options=[
                Option(
                    label="Option 1",
                    action_id="option11"
                ),
                Option(
                    label="Option 2",
                    action_id="option22"
                ),
            ]
        )

        body: Body = Body(
            elements=[
                Container(
                    elements=[
                        select1,
                        select2
                    ]
                )
            ]
        )

        expected = {
            "elements": [
                {
                    "type": "container",
                    "elements": [
                        {
                            "type": "select",
                            "placeholder": "Select an option",
                            "multiselect": False,
                            "trigger_on_input": True,
                            "options": [
                                {
                                    "label": "Option 1",
                                    "action_id": "option1"
                                },
                                {
                                    "label": "Option 2",
                                    "action_id": "option2",
                                }
                            ],
                            'option_groups': []
                        },
                        {
                            "type": "select",
                            "placeholder": "Select an option",
                            "multiselect": False,
                            "trigger_on_input": True,
                            "options": [
                                {
                                    "label": "Option 1",
                                    "action_id": "option11"
                                },
                                {
                                    "label": "Option 2",
                                    "action_id": "option22",
                                }
                            ],
                            'option_groups': []
                        }
                    ]
                }
            ]
        }

        self.assertEqual(expected, body.dict(exclude_none=True))



import os

from switcore.auth.oauth2 import OAuth2
from switcore.auth.utils import get_swit_openapi_base_url


def get_oauth() -> OAuth2:
    swit_client_id: str | None = os.getenv('SWIT_CLIENT_ID', None)
    swit_client_secret: str | None = os.getenv('SWIT_CLIENT_SECRET', None)
    base_url: str | None = os.getenv('BASE_URL', None)
    bot_redirect_url: str | None = os.getenv('BOT_REDIRECT_URL', "/auth/callback/bot")
    user_redirect_url: str | None = os.getenv('USER_REDIRECT_URL', "/auth/callback/user")

    assert swit_client_id is not None, "SWIT_CLIENT_ID is not set check .env file"
    assert swit_client_secret is not None, "SWIT_CLIENT_SECRET is not set check .env file"
    assert base_url is not None, "BASE_URL is not set check .env file"

    return OAuth2(
        client_id=swit_client_id,
        client_secret=swit_client_secret,
        base_url=base_url,
        authorize_endpoint=f"{get_swit_openapi_base_url()}/oauth/authorize",
        access_token_endpoint=f"{get_swit_openapi_base_url()}/oauth/token",
        refresh_token_endpoint=f"{get_swit_openapi_base_url()}/oauth/token",
        bot_redirect_url=bot_redirect_url,
        user_redirect_url=user_redirect_url,
    )



from enum import Enum

from pydantic import root_validator

from switcore.pydantic_base_model import SwitBaseModel


class TagColorTypes(str, Enum):
    primary = "primary"
    secondary = "secondary"
    danger = "danger"


class TagShapeTypes(str, Enum):
    rectangular = "rectangular"
    rounded = "rounded"


class TagStyle(SwitBaseModel):
    color: TagColorTypes = TagColorTypes.secondary
    shape: TagShapeTypes = TagShapeTypes.rectangular

    class Config:
        use_enum_values = True

    @root_validator
    def check_color_and_shape(cls, values):
        color: TagColorTypes | None = values.get('color')
        shape: TagShapeTypes | None = values.get('shape')
        if color is None and shape is None:
            raise ValueError("color and shape cannot be None at the same time")
        return values


class Tag(SwitBaseModel):
    type: str = "tag"
    content: str
    style: TagStyle | None = None


class SubText(SwitBaseModel):
    type: str = "subtext"
    content: str


class OpenOauthPopup(SwitBaseModel):
    action_type: str = "open_oauth_popup"
    link_url: str


class OpenLink(SwitBaseModel):
    action_type: str = "open_link"
    link_url: str


class CloseView(SwitBaseModel):
    action_type: str = "close_view"


class WriteToClipboard(SwitBaseModel):
    action_type: str = "write_to_clipboard"
    content: str


StaticAction = OpenOauthPopup | OpenLink | WriteToClipboard | CloseView



import os
import unittest
from unittest.mock import patch
from urllib.parse import unquote, urlparse, parse_qs

from fastapi import Depends
from httpx import Response
from starlette import status
from starlette.requests import Request
from starlette.responses import Response as StarletteResponse, HTMLResponse, RedirectResponse
from starlette.testclient import TestClient

from switcore.auth.dependencies import get_oauth
from switcore.auth.oauth2 import OAuth2AuthorizeCallbackABC, OAuth2, SecureOAuth2AuthorizeCallbackABC
from switcore.auth.router import router as auth_router
from switcore.auth.schemas import SwitToken
from tests.utils import create_fastapi_app


class OAuth2Test(unittest.TestCase):
    def setUp(self) -> None:
        self.app = create_fastapi_app()
        self.app.include_router(auth_router, prefix="/auth")
        self.client = TestClient(self.app)
        super().setUp()

    def test_bot_authorization_url(self):
        response: Response = self.client.get("/auth/bot", follow_redirects=False)
        expected_url: str = "https://openapi.swit.io/oauth/authorize?" \
                            "response_type=code" \
                            "&client_id=test_client_id" \
                            "&scope=app:install" \
                            "&redirect_uri=test_base_url/auth/callback/bot"
        decoded_url: str = unquote(response.headers["location"])
        self.assertEqual(expected_url, decoded_url)

    def test_oauth2_callback_handler(self):
        expected = {
            "access_token": "eyJhbGciOiJIUzUxMiIsInR5cxxxxxxx.xxxxxxxx.xxxxxxxxxx",
            "expires_in": 604800,
            "refresh_token": "QIDQFWZZXFCXXXXX_XXXXX",
            "scope": "app:install",
            "token_type": "Bearer"
        }

        class BotAuthorizeCallback(OAuth2AuthorizeCallbackABC):
            async def process_access_token(
                    _self,  # noqa F841
                    swit_token: SwitToken,
                    state: str | None
            ) -> StarletteResponse:
                self.assertEqual(swit_token.access_token, expected["access_token"])
                self.assertEqual(swit_token.expires_in, expected["expires_in"])
                self.assertEqual(swit_token.refresh_token, expected["refresh_token"])
                self.assertEqual(swit_token.scope, expected["scope"])
                self.assertEqual(swit_token.token_type, expected["token_type"])
                return HTMLResponse(content="<script>window.close()</script>", status_code=status.HTTP_200_OK)

        @self.app.get("/auth/callback/bot")
        async def bot_oauth_callback(
                request: Request,  # noqa F841
                code: str,
                state: str | None = None,
                oauth: OAuth2 = Depends(get_oauth)
        ):
            callback = BotAuthorizeCallback(
                client=oauth,
                redirect_url="test_base_url/auth/callback/bot",
            )

            return await callback(
                code=code,
                state=state,
            )

        with patch('httpx.AsyncClient.post', return_value=Response(200, json=expected)) as mock_post:
            response = self.client.get("/auth/callback/bot?code=test_code&state=test_state")
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.content, b"<script>window.close()</script>")
            mock_post.assert_called_once()

    def test_oauth2_secure_state(self):
        expected = {
            "access_token": "eyJhbGciOiJIUzUxMiIsInR5cxxxxxxx.xxxxxxxx.xxxxxxxxxx",
            "expires_in": 604800,
            "refresh_token": "QIDQFWZZXFCXXXXX_XXXXX",
            "scope": "app:install",
            "token_type": "Bearer"
        }

        pass_callback_payload = {
            "user_id": "el.lee@swit.io",
            "workspace": "swit_group",
        }

        class SecureAuthorizeCallback(SecureOAuth2AuthorizeCallbackABC):
            async def process_access_token(
                    _self,  # noqa F841
                    swit_token: SwitToken,
                    callback_payload: dict) -> StarletteResponse:
                self.assertEqual(pass_callback_payload["user_id"], callback_payload["user_id"])
                self.assertEqual(pass_callback_payload["workspace"], callback_payload["workspace"])
                self.assertEqual(swit_token.access_token, expected["access_token"])
                self.assertEqual(swit_token.expires_in, expected["expires_in"])
                self.assertEqual(swit_token.refresh_token, expected["refresh_token"])
                self.assertEqual(swit_token.scope, expected["scope"])
                self.assertEqual(swit_token.token_type, expected["token_type"])
                return HTMLResponse(content="<script>window.close()</script>", status_code=status.HTTP_200_OK)

        @self.app.get("/auth/callback/secure-user")
        async def user_oauth_callback(
                request: Request,  # noqa F841
                code: str,
                state: str | None = None,
                oauth: OAuth2 = Depends(get_oauth)
        ):
            callback = SecureAuthorizeCallback(
                client=oauth,
                redirect_url="test_base_url/auth/callback/user",
            )

            return await callback(
                code=code,
                state=state,  # type: ignore
            )

        @self.app.get("/auth/secure-user")
        async def user_authorization(
                request: Request,  # noqa: F841
                oauth: OAuth2 = Depends(get_oauth)
        ):
            scope: str | None = os.getenv('SCOPES', None)

            assert scope is not None, "SCOPES is not set check .env file"

            url: str = await oauth.get_secure_authorization_url(
                redirect_uri=oauth.base_url + "/auth/callback/user",
                scope=scope,
                callback_payload=pass_callback_payload
            )

            return RedirectResponse(
                url,
                status.HTTP_307_TEMPORARY_REDIRECT
            )

        response: Response = self.client.get("/auth/secure-user", follow_redirects=False)

        self.assertEqual(response.status_code, status.HTTP_307_TEMPORARY_REDIRECT)

        parsed_url = urlparse(response.headers.get("Location"))
        query_params = parse_qs(parsed_url.query)

        self.assertEqual(len(query_params['state']), 1)
        jwt_state: str = query_params['state'][0]

        with patch('httpx.AsyncClient.post', return_value=Response(200, json=expected)) as mock_post:
            response = self.client.get(f"/auth/callback/secure-user?code=test_code&state={jwt_state}")
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.content, b"<script>window.close()</script>")
            mock_post.assert_called_once()



from enum import Enum
from typing import Any

from pydantic import root_validator, validator

from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.element_components import Tag
from switcore.ui.image import Image
from switcore.ui.select_item import SelectItem


class NoOptionsReason(SwitBaseModel):
    message: str


class Option(SelectItem):
    """
        https://devdocs.swit.io/docs/user-actions/ref/schemas/select
    """
    icon: Image | None = None
    tag: Tag | None = None


class OptionGroup(SwitBaseModel):
    label: str
    options: list[Option]

    # add validator to check if options is not empty
    @validator('options')
    def options_must_not_be_empty(cls, v):
        if len(v) == 0:
            raise ValueError("options must not be empty")
        return v


class SelectStyleTypes(str, Enum):
    filled = "filled"
    outlined = "outlined"
    ghost = "ghost"


class Style(SwitBaseModel):
    variant: SelectStyleTypes = SelectStyleTypes.outlined


class SelectQuery(SwitBaseModel):
    query_server: bool = True
    disabled: bool = False
    placeholder: str | None = None
    value: str | None = None
    action_id: str | None = None


class Select(SwitBaseModel):
    type: str = 'select'
    placeholder: str | None = None
    multiselect: bool = False
    trigger_on_input: bool = False
    value: list[str] | None = None
    options: list[Option] = []
    option_groups: list[OptionGroup] = []
    no_options_reason: NoOptionsReason | None = None
    style: Style | None = None
    query: SelectQuery | None = None

    @root_validator
    def validate_options_exclusivity(cls, values: dict[str, Any]):
        options: list[Option] = values.get('options', [])
        option_groups: list[OptionGroup] = values.get('option_groups', [])
        if len(options) > 0 and len(option_groups) > 0:
            raise ValueError("Only one of 'options' or 'option_groups' should be provided.")
        return values



from switcore.pydantic_base_model import SwitBaseModel
from switcore.ui.button import Button
from switcore.ui.image import Image
from switcore.ui.select_item import SelectItem


class Header(SwitBaseModel):
    title: str
    subtitle: str | None = None
    context_menu: list[SelectItem] | None = None
    buttons: list[Button] | None = None


class AttachmentHeader(SwitBaseModel):
    title: str
    subtitle: str | None = None
    app_id: str
    icon: Image | None = None



from pydantic import BaseModel


class SwitUser(BaseModel):
    cmp_id: str
    color: str | None
    department: str | None
    id: str
    language: str
    name: str
    photo: str | None
    status: int | None
    status_msg: str | None
    tel: str | None
    timezone: str | None
    user_email: str | None



import unittest

from switcore.action.activity_router import ActivityRouter, PathResolver
from switcore.action.exceptions import UndefinedSubmitAction
from switcore.action.schemas import SwitRequest, BaseState, SwitResponse, ViewCallbackType
from tests.utils import ActivityHandler, create_submit_swit_request


class RouterTest(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self) -> None:
        self.activity_handler = ActivityHandler()
        return await super().asyncSetUp()

    async def test_router01(self):
        activity_router = ActivityRouter()
        action_id: str = str(PathResolver("test_action_id01"))
        swit_request: SwitRequest = create_submit_swit_request("right_panel", action_id)

        @activity_router.register("test_action_id01")
        async def draw_webhook_create(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))

    async def test_router02(self):
        activity_router = ActivityRouter()
        path_resolver = PathResolver("test_action_id01", ["a", 1])
        swit_request: SwitRequest = create_submit_swit_request("right_panel", str(path_resolver))

        @activity_router.register("test_action_id01", ["right_panel"])
        async def draw_webhook_create(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))

    async def test_router03(self):
        activity_router01 = ActivityRouter()
        activity_router02 = ActivityRouter()
        action_id_01: str = str(PathResolver("test_action_id01"))
        action_id_02: str = str(PathResolver("test_action_id02"))

        swit_request_01: SwitRequest = create_submit_swit_request("right_panel01", action_id_01)
        swit_request_02: SwitRequest = create_submit_swit_request("right_panel02", action_id_02)

        @activity_router01.register("test_action_id01", ["right_panel01"])
        async def draw_func_01(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router02.register("test_action_id02", ["right_panel02"])
        async def draw_func_02(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router01)
        self.activity_handler.include_activity_router(activity_router02)

        swit_response_01 = await self.activity_handler.on_turn(swit_request_01, BaseState())
        swit_response_02 = await self.activity_handler.on_turn(swit_request_02, BaseState())

        self.assertTrue(isinstance(swit_response_01, SwitResponse))
        self.assertTrue(isinstance(swit_response_02, SwitResponse))

    async def test_router04(self):
        activity_router = ActivityRouter()
        action_id_01: str = str(PathResolver("test_action_id01"))
        action_id_02: str = str(PathResolver("test_action_id02"))
        action_id_03: str = str(PathResolver("test_action_id03", ["a", 1]))

        swit_request_01: SwitRequest = create_submit_swit_request("right_panel", action_id_01)
        swit_request_02: SwitRequest = create_submit_swit_request("right_panel", action_id_02)
        swit_request_03: SwitRequest = create_submit_swit_request("right_panel", action_id_03)

        invalid_swit_request_01: SwitRequest = create_submit_swit_request("modal", action_id_01)

        @activity_router.register("test_action_id01", ["right_panel"])
        async def draw_func_01(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router.register("test_action_id02", ["right_panel"])
        async def draw_func_02(request: SwitRequest, state: BaseState):  # noqa F811
            return SwitResponse(callback_type=ViewCallbackType.update)

        @activity_router.register("test_action_id03")
        async def draw_func_03(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response_01 = await self.activity_handler.on_turn(swit_request_01, BaseState())
        swit_response_02 = await self.activity_handler.on_turn(swit_request_02, BaseState())
        swit_response_03 = await self.activity_handler.on_turn(swit_request_03, BaseState())

        self.assertTrue(isinstance(swit_response_01, SwitResponse))
        self.assertTrue(isinstance(swit_response_02, SwitResponse))
        self.assertTrue(isinstance(swit_response_03, SwitResponse))

        with self.assertRaises(UndefinedSubmitAction):
            await self.activity_handler.on_turn(invalid_swit_request_01, BaseState())

    async def test_router_with_slash(self):
        activity_router = ActivityRouter()
        swit_request = create_submit_swit_request("right_panel", str(PathResolver("test_action_id01/aa/bb", ["a/a/b", 1])))

        @activity_router.register("test_action_id01/aa/bb", ["right_panel"])
        async def draw_webhook_create(request: SwitRequest, state: BaseState, a: str, number: int):  # noqa F811
            self.assertEqual(a, "a/a/b")
            self.assertEqual(number, 1)
            return SwitResponse(callback_type=ViewCallbackType.update)

        self.activity_handler.include_activity_router(activity_router)

        swit_response = await self.activity_handler.on_turn(swit_request, BaseState())
        self.assertTrue(isinstance(swit_response, SwitResponse))



from switcore.pydantic_base_model import SwitBaseModel


class DatePicker(SwitBaseModel):
    type: str = "datepicker"
    action_id: str
    placeholder: str | None = None
    trigger_on_input: bool = False
    value: str | None = None



from abc import ABC
from collections import defaultdict

from switcore.action.activity_router import ActivityRouter
from switcore.type import AsyncDrawerHandler, DrawerHandler


class BaseHandler(ABC):
    def __init__(self) -> None:
        self._handlers: dict[str, AsyncDrawerHandler | DrawerHandler] = dict()
        self._view_actions: dict[str, set[str]] = defaultdict(set)

    def include_activity_router(self, activity_router: ActivityRouter):
        self._view_actions.update(activity_router.view_actions)
        for action_id, func in activity_router.handlers.items():
            self.add_handler(action_id, func)

    def add_handler(self, action_id: str, func: AsyncDrawerHandler | DrawerHandler):
        self._handlers[action_id] = func

    def get_handler(self, action_id: str) -> AsyncDrawerHandler | DrawerHandler:
        return self._handlers[action_id]



from datetime import datetime

from pydantic import BaseModel


class Workspace(BaseModel):
    admin_ids: list[str] | None
    color: str
    created: datetime
    domain: str
    id: str
    master_id: str
    name: str
    photo: str
